#include "GUIInputStreamIt.h"

extern "C"
{
#include <ctype.h>

#include "r_typedefs.h"
#include "r_rvapi_header.h"
#include "iodefine_cfg.h"
#include "compiler_settings.h"
#include "application_cfg.h"
#include "dev_drv.h"
#include "control.h"
#include "r_devlink_wrapper.h"
#include "command.h"

#include "gpio_iobitmask.h"
#include "rza_io_regrw.h"
#include "r_task_priority.h"
}

#define INVALID_AXIS_VALUE -1

static int_t an_tp_evt_id[6];

// for touch-screen-init
#define TP_INT_PRI (30uL)
#define TP_TSK_PRI (R_OS_TASK_MAIN_TASK_PRI+1)

#include "eC_String.h"
#include "GUIEvent.h"
#include "GUITimer.h"
#include "GUITrace.h"

#include "GUIMemLeakWatcher.h"
#include "WindowsLeakWatcher.h"

#define QUEUED_EVENTS 0

static CGUIInputStreamIt* s_pkInputDevice = NULL;

// callback for touch-events
static void touchevent_callback(int_t nId, TP_TouchEvent_st* psTouchEvt)
{
    if (NULL != s_pkInputDevice)
    {
        s_pkInputDevice->AddEvent(psTouchEvt->sFinger[0]);
    }
}

CGUIInputStreamIt::CGUIInputStreamIt(
    const eC_Int& iTouchScreenX,
    const eC_Int& iTouchScreenY,
    const eC_Int& iTouchScreenWidth,
    const eC_Int& iTouchScreenHeight,
    const eC_Int& iTouchFilterDelta) :
    m_bEventReady(false),
    m_iTouchScreenX(iTouchScreenX),
    m_iTouchScreenY(iTouchScreenY),
    m_iTouchScreenWidth(iTouchScreenWidth),
    m_iTouchScreenHeight(iTouchScreenHeight),
    m_iTouchFilterDelta(iTouchFilterDelta),
    m_iLastDraggedX(INVALID_AXIS_VALUE),
    m_iLastDraggedY(INVALID_AXIS_VALUE),
    m_iLastPositionX(INVALID_AXIS_VALUE),
    m_iLastPositionY(INVALID_AXIS_VALUE)
{
    if (true == m_kSemaphore.Create("Semaphore"))
    {
        GUILOG(GUI_TRACE_DEBUG, "CGUIInputStreamIt: Semaphore created\n");
    }
    else
    {
        GUILOG(GUI_TRACE_DEBUG, "CGUIInputStreamIt: Semaphore NOT created\n");
    }

    /* initialise capacitive touch screen */
    TouchPanel_Init();

    /* open the I2C channel1 driver */
    if (0 > TouchPanel_Open((int_t)m_iTouchScreenWidth, (int_t)m_iTouchScreenHeight, TP_INT_PRI, TP_TSK_PRI, R_OS_ABSTRACTION_PRV_DEFAULT_STACK_SIZE))
    {
        GUILOG(GUI_TRACE_DEBUG, "Error initializing touch-screen\n");
    }
    else
    {
        // register callback function
        an_tp_evt_id[0] = TouchPanel_EventEntry(TPEVT_ENTRY_ALL, m_iTouchScreenX, m_iTouchScreenY, m_iTouchScreenWidth, m_iTouchScreenHeight, &touchevent_callback);
        GUILOG(GUI_TRACE_DEBUG, "touch-screen init successful\n");
    }
    s_pkInputDevice = this;
}

CGUIInputStreamIt::~CGUIInputStreamIt()
{
    // close connection to touch
    TouchPanel_Close();
}

void CGUIInputStreamIt::CreateInstance(
    const eC_Int& iTouchScreenX,
    const eC_Int& iTouchScreenY,
    const eC_Int& iTouchScreenWidth,
    const eC_Int& iTouchScreenHeight,
    const eC_Int& iTouchFilterDelta)
{
    if (!CGUIComponentManager::GetInstance().HasInputMedia())
        CGUIComponentManager::GetInstance().SetInputMedia(new CGUIInputStreamIt(iTouchScreenX, iTouchScreenY, iTouchScreenWidth, iTouchScreenHeight, iTouchFilterDelta));
}

void CGUIInputStreamIt::AddEvent(TP_TouchFinger_st kTouchEvent)
{
    eC_Bool bCreateEvent = false;

    if (
        (m_eLastTouchEvent != kTouchEvent.eState) ||
        (abs(m_iLastPositionX - kTouchEvent.unPosX) > m_iTouchFilterDelta) ||
        (abs(m_iLastPositionY - kTouchEvent.unPosY) > m_iTouchFilterDelta)
        )
    {
        switch (kTouchEvent.eState)
        {
            case TPEVT_ENTRY_UP:
                m_iLastDraggedX = m_iLastDraggedY = INVALID_AXIS_VALUE;
                m_iLastPositionX = m_iLastPositionY = INVALID_AXIS_VALUE;
                break;

            case TPEVT_ENTRY_DOWN:
                m_iLastPositionX = kTouchEvent.unPosX;
                m_iLastPositionY = kTouchEvent.unPosY;
                break;

            case TPEVT_ENTRY_MOVE:
                m_iLastPositionX = kTouchEvent.unPosX;
                m_iLastPositionY = kTouchEvent.unPosY;
                break;

            default:
                break;
        }
        m_eLastTouchEvent = kTouchEvent.eState;
        bCreateEvent = true;
    }

    if (true == bCreateEvent)
    {
        // add event
#if QUEUED_EVENTS == 1
        m_kSemaphore.Enter();
        m_kEventQueue.AddAtEnd(kTouchEvent);
        m_kSemaphore.Leave();
#else
        m_kLastEvent = kTouchEvent;
        m_bEventReady = true;
#endif
        StopIdle();
    }
}

CGUIEvent* CGUIInputStreamIt::GetEvent(eC_UInt uiIdleTime)
{
    CGUIEvent* pkEvent = NULL;
    const eC_UInt uiSleepInterval = 5;
    const eC_UInt uiSleepCycles = uiIdleTime / uiSleepInterval;
    // wait for driver until new event arrives or Idle time reached
    if (uiIdleTime)
    {
        for (eC_UInt uiSleep = 0; uiSleep < uiSleepCycles; ++uiSleep)
        {
            R_OS_TaskSleep(uiSleepInterval);
            
            if (m_bStopIdle)
                break;
#if QUEUED_EVENTS == 1
            if (m_kEventQueue.GetQuantity())
                break;
#else
            if (m_bEventReady)
                break;
#endif
        }
    }

    m_bStopIdle = false;
    try
    {
#if QUEUED_EVENTS == 1
        if (m_kEventQueue.GetQuantity() > 0)
        {
            m_kSemaphore.Enter();

            // Get the last event
            TP_TouchFinger_st kTouchEvent;
            memset(&kTouchEvent, 0, sizeof(TP_TouchFinger_st));
            m_kEventQueue.RemoveFront(kTouchEvent);

            m_kSemaphore.Leave();

            switch (kTouchEvent.eState)
            {
            case TPEVT_ENTRY_UP:
                pkEvent = new CGUIMouseEvent(ET_LBUTTONUP, kTouchEvent.unPosX, kTouchEvent.unPosY);
                break;

            case TPEVT_ENTRY_DOWN:
                pkEvent = new CGUIMouseEvent(ET_LBUTTONDOWN, kTouchEvent.unPosX, kTouchEvent.unPosY, true);
                break;

            case TPEVT_ENTRY_MOVE:
                pkEvent = new CGUIMouseEvent(ET_MOUSEMOVE, kTouchEvent.unPosX, kTouchEvent.unPosY, true, false);
                break;

            default:
                break;
            }
        }
#else
        if (true == m_bEventReady)
        {
            switch (m_kLastEvent.eState)
            {
            case TPEVT_ENTRY_UP:
                pkEvent = new CGUIMouseEvent(ET_LBUTTONUP, m_kLastEvent.unPosX, m_kLastEvent.unPosY);
                break;

            case TPEVT_ENTRY_DOWN:
                pkEvent = new CGUIMouseEvent(ET_LBUTTONDOWN, m_kLastEvent.unPosX, m_kLastEvent.unPosY, true);
                break;

            case TPEVT_ENTRY_MOVE:
                pkEvent = new CGUIMouseEvent(ET_MOUSEMOVE, m_kLastEvent.unPosX, m_kLastEvent.unPosY, true, false);
                break;

            default:
                break;
            }
            m_bEventReady = false;
        }
#endif
    }
    catch (...)
    {
        GUILOG(GUI_TRACE_ERROR, "something went wrong while acquiring event.\n");
    }

    return pkEvent;
}
