#include "GfxEnvVDC0.h"
#include "GUITrace.h"

#include "GfxWrap.h"

extern "C"
{
    #include "r_rvapi_vdc.h"
    #include "r_rvapi_header.h"
    #include "r_display_init.h"

    void VDC_LcdPortSetting(uint32_t param);

    unsigned int ClearDataCache(void* ptr, uint32_t size);

    void CopyBuffer(int x1, int y1, int width, int height);
}

#define GETSTRIDE(width, bpp) (((width * bpp) + 31u) & ~31u)

static volatile int32_t vsync_count = 0;
static int draw_buffer_index = 0;

static void VSyncCallback(vdc_int_type_t int_type)
{
    if (vsync_count > 0)
    {
        vsync_count--;
    }
}

static void Wait_Vsync(const int32_t wait_count)
{
    vsync_count = wait_count;
    while (vsync_count > 0)
    {
        eC_Wait(2);
    }
}

eC_Bool CGfxEnvVDC0::CreateInstance(
    const SurfaceMode_t& eSurfaceMode,
    const CGUICommonEnums::ImageType_t& eScreenFormat,
    const eC_UInt& uiWidth,
    const eC_UInt& uiHeight,
    const eC_UInt& uiBufferCount,
    void** ppkBuffers,
    const eC_Bool& bResizable)
{
    eC_Bool bInstanceCreated = false;
    if (!CGUIComponentManager::GetInstance().HasGfxEnv())
    {
        CGfxEnvVDC0* pkGfxEnv = new CGfxEnvVDC0(
            eSurfaceMode, eScreenFormat,
            uiWidth, uiHeight,
            uiBufferCount, ppkBuffers, bResizable);
        if (pkGfxEnv)
        {
            CGUIComponentManager::GetInstance().SetGfxEnv(pkGfxEnv);
            bInstanceCreated = true;
        }
    }
    return bInstanceCreated;
}

CGfxEnvVDC0::CGfxEnvVDC0(
    const SurfaceMode_t& eSurfaceMode,
    const CGUICommonEnums::ImageType_t& eScreenFormat,
    const eC_UInt& uiWidth,
    const eC_UInt& uiHeight,
    const eC_UInt& uiBufferCount,
    void** ppkBuffers,
    const eC_Bool& bResizable) :
    CGfxEnv(eSurfaceMode, eScreenFormat, uiWidth, uiHeight, uiBufferCount, ppkBuffers, bResizable)
{
    // setup vdc
    vdc_error_t error;
    vdc_channel_t vdc_ch = VDC_CHANNEL_0;

    /***********************************************************************/
    /* display init (VDC output setting) */
    /***********************************************************************/
    error = r_display_init (vdc_ch);
    if (error == VDC_OK)
    {
        error = R_RVAPI_InterruptEnableVDC(vdc_ch, VDC_INT_TYPE_S0_LO_VSYNC, 0, VSyncCallback);
    }

    /* Image Quality Adjustment */
    if (VDC_OK == error)
    {
        error = r_image_quality_adjustment(vdc_ch);
    }

    /* Enable signal output */
    if (VDC_OK == error)
    {
        /* Wait for register update */
        eC_Wait(16);

        R_RVAPI_DispPortSettingVDC(vdc_ch, &VDC_LcdPortSetting);
    }
    
    /* we support merge */
    m_bNeedsManualMerge = false;
    
    m_bSupportsCopy = true;
    
    SetMainLayerProperties(VDC_LAYER_ID_0_RD, true);
}

CGfxEnvVDC0::~CGfxEnvVDC0()
{
}

void CGfxEnvVDC0::InitLayer(const LayerConfiguration_t& kLayerConfig)
{
    vdc_error_t error;
    vdc_channel_t vdc_ch = VDC_CHANNEL_0;

    gr_surface_disp_config_t gr_disp_cnf;
    uint32_t  clut_table[4] = {
            0x00000000, /* No.0 transparent color  */
            0xFF000000, /* No.1 black */
            0xFF00FF00, /* No.2 green */
            0xFFFF0000  /* No.3 red */
    };

    /* buffer clear */
    for (unsigned int uiBufferIndex = 0; uiBufferIndex < kLayerConfig.m_uiNumBuffers; ++uiBufferIndex)
    {
        memset((void*)kLayerConfig.m_pkBuffers[uiBufferIndex], 0, kLayerConfig.m_uiPaddedWidth * kLayerConfig.m_uiHeight * GetBytesPerPixel(kLayerConfig.m_ePixFormat));
    }

    gr_disp_cnf.layer_id = (vdc_layer_id_t)kLayerConfig.m_uiLayerID;
    gr_disp_cnf.disp_area.hs_rel = kLayerConfig.m_iXPos;
    gr_disp_cnf.disp_area.hw_rel = kLayerConfig.m_uiWidth;
    gr_disp_cnf.disp_area.vs_rel = kLayerConfig.m_iYPos;
    gr_disp_cnf.disp_area.vw_rel = kLayerConfig.m_uiHeight;
    gr_disp_cnf.fb_buff = (void*)kLayerConfig.m_pkBuffers[0];
    gr_disp_cnf.fb_stride = kLayerConfig.m_uiPaddedWidth * GetBytesPerPixel(kLayerConfig.m_ePixFormat);

    if (GetBytesPerPixel(kLayerConfig.m_ePixFormat) == 2)
    {
        gr_disp_cnf.read_format = VDC_GR_FORMAT_RGB565;
        gr_disp_cnf.read_swap = VDC_WR_RD_WRSWA_32_16BIT;
    }
    else if (GetBytesPerPixel(kLayerConfig.m_ePixFormat) == 4)
    {
        gr_disp_cnf.read_format = VDC_GR_FORMAT_RGB888;
        gr_disp_cnf.read_swap = VDC_WR_RD_WRSWA_32BIT;
    }

    gr_disp_cnf.clut_table = clut_table;
    gr_disp_cnf.read_ycc_swap = VDC_GR_YCCSWAP_CBY0CRY1; // !!!
    if (kLayerConfig.m_uiLayerID == (int)VDC_LAYER_ID_0_RD)
        gr_disp_cnf.disp_mode = VDC_DISPSEL_CURRENT;
    else
        gr_disp_cnf.disp_mode = VDC_DISPSEL_BLEND;

    R_RVAPI_GraphCreateSurfaceVDC(vdc_ch, &gr_disp_cnf);
}

void CGfxEnvVDC0::EndOfRefresh()
{
    void* fb_buff = (void *)m_akLayers[0].m_pkDisplayBuffer;
    uint32_t stride = GETSTRIDE(m_akLayers[0].m_uiPaddedWidth, GetBytesPerPixel(m_akLayers[0].m_ePixFormat));
    uint32_t height =  m_akLayers[0].m_uiHeight;

    unsigned int cache_error = ClearDataCache(fb_buff, stride * height);
    if (cache_error)
    {
        GUILOG(GUI_TRACE_ERROR, "error ClearDataCache.\n");
    }

    vdc_error_t error = R_RVAPI_GraphChangeSurfaceVDC(VDC_CHANNEL_0, VDC_LAYER_ID_0_RD, (void*)fb_buff);
    if (VDC_OK == error)
    {
         Wait_Vsync(1);
    }
}

eC_Bool CGfxEnvVDC0::RefreshScreen(const CGUIRect& rkInvalidatedArea)
{
    if (m_bSupportsCopy)
    {
        eC_Int iX1 = eC_ToInt(eC_Floor(rkInvalidatedArea.GetX1()));
        eC_Int iY1 = eC_ToInt(eC_Floor(rkInvalidatedArea.GetY1()));
        eC_Int iX2 = eC_ToInt(eC_Ceil(rkInvalidatedArea.GetX2()));
        eC_Int iY2 = eC_ToInt(eC_Ceil(rkInvalidatedArea.GetY2()));
        eC_Int iRectWidth = iX2 - iX1 + 1;
        eC_Int iRectHeight = iY2 - iY1 + 1;
        
        if (iX1 < 0)
        {
            iRectWidth += iX1;
            iX1 = 0;
        }
        if (iY1 < 0)
        {
            iRectHeight += iY1;
            iY1 = 0;
        }

        if (iX1 + iRectWidth > m_uiWidth)
        {
            iRectWidth = m_uiWidth - iX1;
        }
        if (iY1 + iRectHeight > m_uiHeight)
        {
            iRectHeight = m_uiHeight - iY1;
        }
        
        if ((iRectWidth > 0) && (iRectHeight > 0))
        {
            CopyBuffer(iX1, iY1, iRectWidth, iRectHeight);
        }
    }
    return true;
}

void CGfxEnvVDC0::CopyBuffer(int iX1, int iY1, int iRectWidth, int iRectHeight)
{
    eC_UInt uiBytesPerPixel = GetBytesPerPixel(m_akLayers[0].m_ePixFormat);
    uint8_t* pubSource = (uint8_t*)m_akLayers[0].m_pkDisplayBuffer + (iY1 * m_uiWidth + iX1) * uiBytesPerPixel;
    uint8_t* pubDest = (uint8_t*)m_akLayers[0].m_pkRenderBuffer + (iY1 * m_uiWidth + iX1) * uiBytesPerPixel;
    uint32_t pitch = m_uiWidth * uiBytesPerPixel;
    for (uint16_t iLineIndex = 0; iLineIndex < iRectHeight; ++iLineIndex)
    {
        memcpy(pubDest, pubSource, iRectWidth * uiBytesPerPixel);

        pubSource += pitch;
        pubDest += pitch;
    }
}
